#!/usr/bin/perl
# | < >
# Time flies like an arrow
# (Fruit flies like a banana)

# Programm fr die automatische Kategorisierung von Mailinglisten (Inhaltsanalyse)
# (c) 2004 Kajetan Hinner und   Univ. Mainz und Univ. Frankfurt, Deutschland
#
# aktuelle Version unter www.hinner.com/programs
#
# das Programm erwartet im store-Modus die bergabe des Dateinamens des ML-Logs auf der Kommandozeile
# und schreibt die Daten in eine Datenbank
# fr jede gefundene E-Mail wird ein Eintrag angelegt
# Benutzung:
# mlanalysis -h
#
# to dump tables: mysqldump -u root -d kdeml1 > /tmp/z
# to create new db: (sql shell): CREATE DATABSE gnomedevel;
# to insert tables: mysql -u root gnomedevel < /tmp/z

#
# NIE vergessen bei perl! Bei foreach $variable (@array) wird $variable
# immer absolut lokal und nicht weitergegeben. Ich wei nicht, ob 
# das ein Bug oder Feature sein soll.

# "KDE" - KDE-Modus: KDE Mailinglisten
# KDE-Spool-Modus: KDE-Mailinglisten (nicht ber ML-Software, sondern selbst gespeichert)
# "Gnome" - Gnome-Modus: Gnome-Mailingslisten

# Globaler Modus (unbedingt vor der Analyse hier einstellen)

use strict;
use warnings;

use Date::Parse; 	# Module to transform Date
use MIME::WordDecoder;
# my $wdecode = supported MIME::WordDecoder "ISO-8859-1";	# get a Word-Decoder

my $dbh;			  # data base handle
my $sql_command; 		  # will be used to give commands to db
my $sql_sth;			  # stamentent handle
my $sql_rv;			  # return value

my $mlarchivefile;		  # filename of ml-file
my @filenamelist=();		  # all filenames to process (command line *)
my $edgelistfile="edge.list";	  # filename of edgelist-file
my $mla_mode;			  # "store, pass2, buildlists"
				  # mode for program
my $mla_startdate="1980-01-31";	  # begin of investigation
my $mla_enddate="2008-12-31";	  # end of investigation

 


my $scanmode = "KDE"; # Mails getrennt mit From / From:

my @mlisolate_rexp;	# Regexp to check for certain content (isolate E-Mails of Mailing List file, beginning and end)
my $mlisolate_ifnum; # number of "ifs" to match

# Einstellungen je nach Scan-Modus

if ($scanmode eq "KDE") 
  {
    $mlisolate_rexp[0]="^From";
    $mlisolate_rexp[1]="^From:";
    $mlisolate_ifnum = 2;
  }

if ($scanmode eq "GNOME") 
  {
    $mlisolate_rexp[0]='^From\s';
    $mlisolate_rexp[1]='^Received:\s'; # works only until April 2000
    # $mlisolate_rexp[1]='^Return-Path:\s'; # from May 2000 on
    $mlisolate_ifnum = 2;
  }


#
# Subroutinen
#


sub get_options {

my $rst_argv = @ARGV;
my $rst_arg;

if ($rst_argv == 0)
 {
  return (-1);  # quit, no command line options
 }
 
  else
  
   {
    print ("\nCommand line arguments: $rst_argv\n");

    foreach (@ARGV)
     {
      print ("Argument: '$_'\n");

      if (/^-h/) { return(-1);} # print help and exit
      
      if (/^\-store$/)    # store_mode
        {
	 $mla_mode = "store";
  	}
	elsif (/^-pass2$/)
        {
	 $mla_mode = "pass2";
  	}
	elsif (/^-buildedgelist$/)
        {
	 $mla_mode = "buildedgelist";
	}
	elsif (/^-startdate=(.*)$/)
        {
	 $mla_startdate = $1;
	}
	elsif (/^-enddate=(.*)$/)
        {
	 $mla_enddate = $1;
	}
  
	# letzter Paramter, einfach solo
	elsif ($mla_mode eq "store")
        {
 	 # print ("Speichere Dateinamen '$_'\n");
	 push (@filenamelist, $_);	# store filenames	 
	}
	elsif ($mla_mode eq "buildedgelist")
        {
	 $edgelistfile=$_;
	}	
	else {
	 print ("Unknown paramter: '$_'\n");
	 return (-1);
	 }
       } # end for each
   } # end if there are command line options
 
 return (0);
 
 } # end sub get_options	
 
 
 
 
# Schleife ber die Datei

# MLARCHIVEFILE: File Handle fr die Datei, die alle E-Mails der Mailingliste summarisch enthlt
# $mlarchivefilelineno : Zeilennummer, in der sich das Script beim Einlesen dieser Datei gerade befindet
# $mlarchivefileline: Inhalt der Zeile (Text), die gerade abgearbeitet wird
# @mlarchivemail: E-Mail der Mailingliste, die gerade behandelt wird
# @mlarchiveprefix: Anfang der neuen E-Mail (was zuviel gelesen wurde)

my @mlarchiveprefix=();		# reset saved mail content
my @mlarchivemail;

my $mlarchivefilelineno;	# reset line number of line to read
my $mlarchivefileEOF=0;

my $mlifcounter;  # zhlt, wie viele Bedingungen beim Parsen bereits erfllt sind


my $message_id;		# E-Mail-ID of currently processed mail
my $message_date;	# E-Mail-Date of currently processed mail
my $message_date_ts;	# E-Mail-Date, Unix timestamp format
my $message_inreplyto;	# EMail In-Reply-To field
my $message_first;	# 1/0: first message of thread
my $message_subject;	# EMail Subject
my @mail_body;		# Mail text (body)
my @mail_header;	# Mail header
my @mail_references;	# hash with all mail references (one per line)
my $autorennamen_autorid;	# ID in table Autorennamen
my $autor_id;		# ID aus der Autor-Tabelle

my $date_calc;		# For Date and time calulations/transformations (Short storage)


sub get_email {

   my $matchregexpcounter = 0; # Zhler zurcksetzen
   $mlifcounter =0;

   print ("\n---------SEARCHING FOR NEW E-MAIL-----------------------------------------------\n");
      
   if (scalar @mlarchiveprefix > 1)
     {
      @mlarchivemail=@mlarchiveprefix;	# new mail begins with surplus read of last E-Mail
      
#      print ("ist grer, nmlich \n". scalar @mlarchiveprefix .  "\n");
#      print ("und die erste Zeile: '$mlarchiveprefix[0]'\n");
# print ("und die zweite Zeile: '$mlarchiveprefix[1]'\n");
# print ("und die dritte Zeile: '$mlarchiveprefix[2]'\n");
 
      
      
      
      @mlarchiveprefix=(); # Mail lschen
     }
   
   my $mlarchivefileline;	# line of mail to read

 GETLINE: while (defined ($mlarchivefileline = <MLARCHIVEFILE> ))
 
  {

    $_ = $mlarchivefileline;

    # print ("Zeile $mlarchivefilelineno: '$mlarchivefileline'\n");

    push (@mlarchivemail, $_); # neu eingelesene Zeile der E-Mail hinzufgen
    # print ("und jetzt wird getestet! (Zeile $mlarchivefilelineno), Vergleichsversuch $mlifcounter, Vergleich '$mlisolate_rexp[$mlifcounter]'\n");
    if (/$mlisolate_rexp[$mlifcounter]/)
     {
 print ("Bedingung erfllt: $mlisolate_rexp[$mlifcounter] in Zeile $mlarchivefilelineno '$_'\n");
       $mlifcounter++;
       
       if ($mlifcounter == $mlisolate_ifnum)
         {
	  if ($mlarchivefilelineno != 2)	# nicht am Dateianfang (fngt immer mit einer neuen E-Mail an)
	    {
	     push @mlarchiveprefix, pop (@mlarchivemail); # remove matched line and add it for the next one
	     push @mlarchiveprefix, pop (@mlarchivemail);

	     $mlarchivefilelineno++;
 	     
	     return();
	    }
	  $mlifcounter=0;	# we are in line 2. reset ifcounter (first mail of File) and read the next lines
	 } 
      
      
      }
     elsif ($mlifcounter) {
      $mlifcounter =0;	# if there is one line without match, restart
      }
 $mlarchivefilelineno++;
 
 } # Ende des whiles
 
#else {last GETLINE}	# end of file

#    return get_email  (0);
#   }
# 
# 
# unshift @mlarchivemail, $fromspaceline;
# unshift @mlarchivemail, $fromcolonline;
#  return get_email (0); # am Ende der Datei angekommen, E-Mail zhlt
 
$mlarchivefileEOF=1;

} # Ende sub get_mail


# set "global" Variables (arrays for Author information)

my @email_autor_fullname;
my @email_autor_email_addresses;

sub get_Autor {

my @AutorName;
my @AutorEMail;
my $name;
my $email;
my @EMail = @_;

# print ("wurde das bergeben? \n@EMail");

# print ("Zeile 1 '$EMail[0]'\n");
# print ("Zeile 2 '$EMail[1]'\n");
# print ("Zeile 3 '$EMail[2]'\n");

# wir konzentrieren uns auf die ersten beiden Zeilen

foreach (@EMail)
 {
   # print ("Autor_Name: Durchsucht '$_'\n");
   
   if (/^From\s/)
    {
#     ($email) = /^From (.+) [Mon|Tue|Wed|Thu|Fri|Sat|Sun] /;
     ($email) =  /^From\s(.+)\s(Mon|Tue|Wed|Thu|Fri|Sat|Sun)\s/;
     if ($email) {
     print ("Yeah, unter 'From ': $email\n");
     push (@AutorEMail, $email);}
    }

   if (/^From:/) {
    if ($scanmode eq "KDE")  {
     
     ($email, $name) = /^From: ([^\(]+)\((.+)\)/;	# regexp: match everything but ( for mail address and name in (...)  
    }
    elsif ($scanmode eq "GNOME") {
     ($name, $email) = /^From: ([^<]+)\s<(.+)\>/;	# regexp: match everything but ( for mail address and name in (...)
     }
     if ($email || $name) {
        print ("Yeah, gefunden unter 'From: ': '$email' und '$name'\n");
        if ($email) {push (@AutorEMail, $email);}
        if ($name) {push (@AutorName, $name);}
	}
    }  

} # end foreach


# add new information to "global" variables
@email_autor_fullname = @AutorName;
@email_autor_email_addresses = @AutorEMail;
return ();

} # end Sub

# get ID of E-Mail and return it (pass E-Mail as Array)
sub get_message_id {

 my @EMail = @_;
 
 foreach (@EMail) {
  if (/^Message-ID: <(.*)?>/i)
    {
     # print ("Message-ID='$1'\n");
     return ("<$1>");
    }
 }
 
 print ("Message-ID nicht gefunden in E-Mail '@EMail'\n");
 
} 

# get In-Reply-To of E-Mail and return it (pass E-Mail as Array)
sub get_message_inreplyto {

 my @EMail = @_;
 
 foreach (@EMail) {
  if (/^In-Reply-To:\s<(.*)?>/)
    {
     # print ("In-Reply-To='$1'\n");
     return ("<$1>");
    }
 }
 
 # print ("In-Reply-To nicht gefunden in E-Mail '@EMail'\n");
 
} 



# get Subject of E-Mail and return it (pass E-Mail as Array)
sub get_message_subject {

 my @EMail = @_;
 
 foreach (@EMail) {
  if (/^Subject:\s+(.*)$/)
    {
     print ("Subject='$1'\n");
     return ("$1");
    }
 }
 
 print ("Subject nicht gefunden in E-Mail '@EMail'\n");
 
} 


# get any date like value from the header field (if Date: does not work)
sub get_any_date {

 my @header = @_;
 
 foreach (@header) {
  if (/((Mon|Tue|Wed|Thu|Fri|Sat|Sun).*(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dez).*\d\d:\d\d:\d\d.*$)/i)
    {
     print ("Datum='$1'\n");
     return ("$1");
    }
 }
 
 print ("Kein Datum gefunden in '@header'\n");
 return (0);
} # end sub get any date



# get Subject of E-Mail and return it (pass E-Mail as Array)
sub get_message_from {

 my @EMail = @_;
 
 foreach (@EMail) {
  if (/^From:\s+(.*)$/)
    {
     print ("From='$1'\n");
     return ("$1");
    }
 }
 
 print ("From nicht gefunden in E-Mail '@EMail'\n");
 
} 


# get Body of E-Mail and return it (pass E-Mail as Array)
# also assign @mail_header
sub get_message_body {

# my @EMail = @_;
 
 my $counter=0;
 my $length = scalar @_;

 if (!$length) {
     @mail_header="";
     return ("");
     print ("Checken: Aufruf von get_message_body mit leerem Array (Zeile: $mlarchivefilelineno)\n");
     }
 
 $length--;
 
 while (!($_[$counter] =~ /^$/)) {
      $counter++;			# search for first empty line
      #print ("Index eins nach oben\n");
}

 #print ("und die erste Zeile: '$_[0]'\n");
 #print ("und die zweite Zeile: '$_[1]'\n");
 #print ("und die dritte Zeile: '$_[2]'\n");
 
 @mail_header = @_[0 .. $counter-1];
 
 print ("Length und Counter: $length; $counter und header ist: '@mail_header'\n");
 
  if ($counter >= $length) {
   return ("");			# empty mail (only header)
   }
   else {
      return (@_[$counter+1 .. $length]);
    }
 }
 
     
# get Date and Time of E-Mail and return it in MyAQL format (pass E-Mail as Array)
sub get_message_date {

 my @EMail = @_;
 
 foreach (@EMail) {
  if (/^Date: (.*)$/)
    {
     # print ("Date='$1'\n");
     # ($ss,$mm,$hh,$day,$month,$year,$zone) = strptime($date);
     
     return ("$1");
    }
 }
 
 print ("Kein Datumseintrag gefunden in E-Mail '@EMail'\n");
 
} 

# checks if this E-Mail is the first one (of a thread), set mail_first to 1 and store references in @mail_references, otherwise 0
sub get_message_references {

 my @EMail = @_;
 # my @tmp_references="";
 @mail_references=();
 my $ref_mode=0;
 
 foreach (@EMail) {
  # if (!$ref_mode && /^References:\s+(\S*)\s*$/)   *** References checken falls zweizeilig
  
  if (!$ref_mode && /^References:\s+(.+)$/)    
    {
     print ("Ersten Verweis gefunden (References): '$1'\n");
     @mail_references = split (/\s+/, $1);
     $ref_mode++;
     # push (@tmp_references, $1);
    }
    
    elsif ($ref_mode) {
     if (/:/) {			# new Header (because :)
       # references einzeln in Array ablegen
       foreach (@mail_references) {
           print ("Verweis gefunden (References): '$_'\n");
         }
        return();	# return to caller
       }
	
	else {
	 print ("GNRK weiteren Verweis gefunden (References): '$_'\n");
         s/^\s+//;		# remove heading white spaces
         s/\s+$//;		# remove trailing white spaces

	 # @tmp_references = split (/\s+/);
	 # push (@mail_references, @tmp_references);
	 push (@mail_references, split (/\s+/));
	 }
     }
 }
 
 if (scalar @mail_references == 0) {
   print ("References nicht gefunden in E-Mail '@EMail'\n");
 }
 
} 




# we need to check the array @email_autor_email_addresses
# if there are duplicate emails in it - they will be delated
# if there is " at " in it - which will be replaced with @

sub check_email_addresses {
my %temp = ();
foreach (@email_autor_email_addresses)
  {
  print ("E-Mail-Adresse: $_\n");
  s/ at /@/;		# recreate "real" E-Mail-Addresses
  s/^\s+//;		# remove heading white spaces
  s/\s+$//;		# remove trailing white spaces
  $temp{$_}=1;
  }
@email_autor_email_addresses = keys %temp;

foreach (@email_autor_email_addresses)
  {
  print ("E-Mail-Adresse: '$_'\n");
  }

}


# check if Author already in "Autorennamen" table - if not, insert him
# return "AutorID"

sub check_Autor {

my $fullnameref;
my $emailref;

my $autorname;		# Name of E-Mail author
my $autor_alias; 	# alias-name (decoded MIME string of author name)
my $sql_autorname;	# for inserting into DB - quoted name
my $sql_sth;
my $sql_rv;
my $if_autor;		# Bedingung fr SQL-Abfrage bzgl. Autor
my $if_email;		# Prfbedingung fr E-Mail
my @db_data;		# row of DataBase

$fullnameref=$_[0];
$emailref=$_[1];

my @fullnames=@$fullnameref;
my @emails=@$emailref;

print ("Ich prfe, ob es den Autor/die Autoren @fullnames mit den E-Mail-Adressen @emails schon gibt\n");

if (scalar @fullnames > 1) {
  $if_autor=join (" or ", @fullnames);
  } else {
   $if_autor=$fullnames[0];
   }
   
print ("Prfbedingung bei check_autor: $if_autor\n");

# $sql_command = "select id, autorid from Autorennamen where autoralias=$sql_autorname";
$sql_command="select autorid from Autorennamen where autorname = ?";
$sql_sth = $dbh->prepare($sql_command);
$sql_rv = $sql_sth->execute($if_autor);
if ($sql_sth->rows == 0)
  {
   print ("Den Autor gibt es unter diesem Namen '$if_autor' noch nicht! Prfe nun die E-Mail-Adresse(n)\n");
   if (scalar @emails > 1) {
    $if_email=join (" or ", @emails);
    } else {
      $if_email=$emails[0];
     }
   $sql_command="select autorid from Autorennamen where autoremail = ?";
   $sql_sth = $dbh->prepare($sql_command);
   $sql_rv = $sql_sth->execute($if_email);
   		# ignore E-Mail address of lists: kde-devel@kde.org" (hack)
   if (($sql_sth->rows > 0) && ($if_email ne 'kde-devel@kde.org')) {
     print ("Der Autor existiert unter einer anderen E-Mail-Adresse\n");
     $autorennamen_autorid=$sql_sth->fetchrow;
     print ("Aus der Autorennamen-Tabelle kam bei '$if_autor': $autorennamen_autorid\n");
     	# add aliasname in table
     $sql_command="UPDATE Autorennamen SET autoralias=CONCAT_WS(',',autoralias, ?) WHERE (autorid=$autorennamen_autorid)";
     # (CONCAT_WS (',',autoralias,?)) 
     
   print ("Update-Statement Alias '$sql_command'\n");
   $sql_sth = $dbh->prepare($sql_command);
   $sql_rv = $sql_sth->execute ($if_autor);
   $sql_sth->finish;
   }
   else {
   
   print ("Eingefgt in die Autorennamen-DB wird '$if_autor'\n");
   # $sql_command = "select id, autorid from Autorennamen where autoralias=$sql_autorname";
   $sql_command="select MAX(autorid) from Autorennamen";
   $sql_sth = $dbh->prepare($sql_command);
   $sql_rv = $sql_sth->execute;
   $autorennamen_autorid=$sql_sth->fetchrow;
   # print ("Alte AutorID ist $autorennamen_autorid\n");
   $autorennamen_autorid += 1;
   print ("Neue AutorID ist $autorennamen_autorid\n");
   
   if ($if_autor =~ /^\=\?iso/i) {
        $autor_alias = unmime ($if_autor);
	}
	else {
            $autor_alias =""; }
   
   $sql_command="insert into Autorennamen (autorid, autorname, autoralias, autoremail) VALUES (?, ?, ?, ?)";
   print ("Statement '$sql_command'\n");
   $sql_sth = $dbh->prepare($sql_command);
   $sql_rv = $sql_sth->execute ($autorennamen_autorid, $if_autor, $autor_alias, $email_autor_email_addresses[0]);
   $sql_sth->finish;
   
   
   }
  }
  else {
   $autorennamen_autorid=$sql_sth->fetchrow;
   print ("Aus der Autorennamen-Tabelle kam bei '$if_autor': $autorennamen_autorid\n");
  }
$sql_sth->finish;









#print ("mail_name1: $mail_name1\n");
 
###   $sq_mail_name1=$dbh->quote($mail_name1);

###   $sql_command="insert into Mail set Subject=$sq_mail_name1";

#   print ("Statement '$sql_command'\n");
 
###  $sth = $dbh->prepare($sql_command) || die "Kann Statement '$sql_command' nicht vorbereiten $DBI::errstr\n";
###  $rv = $sth->execute or die "Kann die Anweisung $sql_command nicht ausfhren\n";
###  $sth->finish;




}


# this sub sets the new date of author in the "Autor" table.
sub set_Autor {

$sql_command="select id from Autor where autorennamenid = ?";
$sql_sth = $dbh->prepare($sql_command);
$sql_rv = $sql_sth->execute($autorennamen_autorid);
if ($sql_sth->rows == 0)
  {
   print ("Den Autor gibt es noch nicht in der Autor-Tabelle!\n");
   print ("Eingefgt wird die ID $autorennamen_autorid\n");
   # $sql_command = "select id, autorid from Autorennamen where autoralias=$sql_autorname";
   
   my $date_calc=str2time($message_date);
   # print ("Ich mchte umwandeln: '$message_date' und herauskam: '$date_calc'\n");
   $sql_command="insert into Autor (autorennamenid, datumerstemail) VALUES ($autorennamen_autorid, FROM_UNIXTIME($date_calc))";
   print ("Statement '$sql_command'\n");
   $sql_sth = $dbh->prepare($sql_command);
 
   $sql_rv = $sql_sth->execute ();
   $sql_sth->finish;
  }
  else {
   $autor_id=$sql_sth->fetchrow;
   print ("Aus der Autor-Tabelle kam als ID: $autor_id\n");
  }
$sql_sth->finish;

my $date_calc=str2time($message_date);
# print ("Ich mchte umwandeln: '$message_date' und herauskam: '$date_calc'\n");

$sql_command="UPDATE Autor SET emailanzahl=emailanzahl+1, ersteranzahl=ersteranzahl+$message_first, datumletztemail=FROM_UNIXTIME($date_calc) WHERE (autorennamenid=$autorennamen_autorid)";
print ("Statement '$sql_command'\n");
$sql_sth = $dbh->prepare($sql_command);
$sql_rv = $sql_sth->execute();
$sql_sth->finish;


}


# this sub sets the new data in the "Mail" table.
sub set_Mail {

my $sql_sth;
my $sql_rv;

$sql_command="select id from Autor where autorennamenid = ?";
$sql_sth = $dbh->prepare($sql_command);
$sql_rv = $sql_sth->execute($autorennamen_autorid);
if ($sql_sth->rows == 0)
  {
   print ("Den Autor '$autorennamen_autorid' gibt es noch nicht bei SetMail() in der Autor-Tabelle!\n");
   print ("Das sollte eigentlich nicht sein. Ich bin enttuscht\n");
   die();
   # $sql_command = "select id, autorid from Autorennamen where autoralias=$sql_autorname";
  }
  else {
   $autor_id=$sql_sth->fetchrow;
   print ("Aus der Autor-Tabelle kam als ID: $autor_id\n");
  }
$sql_sth->finish;

my $date_calc=str2time($message_date);
#print ("Ich mchte umwandeln: '$message_date' und herauskam: '$date_calc'\n");

#$sql_command="insert into Mail (autorennamenid, maildate, messageid ) VALUES ($autorennamen_autorid, FROM_UNIXTIME($message_date_ts), '$message_id')";
#print ("Statement '$sql_command'\n");
#$sql_sth = $dbh->prepare($sql_command);
#$sql_rv = $sql_sth->execute();
#$sql_sth->finish;

$sql_command="insert into Mail (autorennamenid, maildate, messageid, inreplytoid, subject, firstmessage, mailreferences, lengthlines, bodytext ) VALUES ($autorennamen_autorid, FROM_UNIXTIME($message_date_ts), ?, ?, ?, $message_first, ?, ?, ?)";
# print ("Statement '$sql_command'\n");
$sql_sth = $dbh->prepare($sql_command);

$sql_rv = $sql_sth->execute ($message_id, $message_inreplyto, $message_subject, join ("\n", @mail_references), scalar @mail_body, join ("", @mail_body));
$sql_sth->finish;

# print ("ach ja, und der Mail_body ist '@mail_body'\n");


#$sql_command="insert into Mail (autorennamenid, maildate, messageid, inreplytoid, *references, subject, lengthchars, lengthlines, bodytext, firstmessage ) VALUES ($autorennamen_autorid, FROM_UNIXTIME($message_date_ts), '$message_id', '$message_inreplyto', ?, '$message_subject', ?, ?, ?, $message_first)";
#print ("Statement '$sql_command'\n");
#$sql_sth = $dbh->prepare($sql_command);
#$sql_rv = $sql_sth->execute($mail_references[0], length (@mail_body), scalar @mail_body, $mail_body[5]);
#$sql_sth->finish;

}

sub mla_store {
# $mlarchivefile=@_;

$mlarchivefileEOF = 0;		# Reset EOF flag
$mlarchivefilelineno =1;	# start from line 0

print ("\n\nffne Mailinglisten-Datei '$mlarchivefile'\n");  
open (MLARCHIVEFILE, "<$mlarchivefile") or die ("Konnte die Mailinglisten-Archivdatei '$mlarchivefile' nicht ffnen\n");

my $zeilennummer=0;

@email_autor_fullname = "";		# current E-Mail: All possible names of author
@email_autor_email_addresses = "";  	# current E-Mail: All possible E-Mail-addresses of one author

while (!$mlarchivefileEOF) {
# while ($mlarchivefilelineno < 100) {

 &get_email();

 print ("\n\nMail gefunden bis Zeile ", $mlarchivefilelineno-3, "\n");

    # Mail gefunden: Analysieren  ######################

 #  print ("Mail: \n");
 #  print @mlarchivemail;
 #  print "\n";
 
	# get mail body and set @mail_header (in the sub!)
 @mail_body = &get_message_body (@mlarchivemail);
 
  
 # @email_autor_fullname and @email_autor_email_addresses will be set/extended with the following function
 
 $message_id = &get_message_id (@mail_header);
 $message_date = &get_message_date (@mail_header);	# get Date of Mail
 $message_date_ts=str2time($message_date);	# convert to timestamp  format
 if (!$message_date_ts) { # something went wrong with the date conversion
    print ("Ich wollte umwandeln: '$message_date'\n");
    if ($message_date =~ /^\=\?iso/i) {
        $message_date = unmime ($message_date);
	$message_date_ts=str2time($message_date);	
	}
    if (!$message_date_ts) { # something we
       # $message_date =~ s/\s(\S+$)//;	#	 hack: remove timezone code
       $message_date = &get_any_date (@mail_header);
       }
    print ("Ich mchte jetzt umwandeln: '$message_date'\n");
    $message_date_ts=str2time($message_date);	# convert to timestamp  
    if (!$message_date_ts) {
       print ("Cannot decode timezone: Mail $message_id in file $mlarchivefile\n");
       die();
       }
     print ("Special decode timezone: Mail $message_id in file $mlarchivefile - $message_date_ts\n");
    } 
 

# if ($scanmode eq "GNOME") 
#  {
#   $message_id = &get_message_from (@mail_header);  
#  }
    
 $message_inreplyto = &get_message_inreplyto (@mail_header);
 
 &get_Autor (@mail_header);  
 print ("Gefunden wurde an Autornamen: @email_autor_fullname und an zugehrigen E-Mail-Adressen: @email_autor_email_addresses \n");
 
 &check_email_addresses ();	# check @email_autor_email_addresses
 
 &check_Autor (\@email_autor_fullname, \@email_autor_email_addresses);	# prft, ob es diesen Autor in der DB schon gibt und setzt die Variable $autorennamen_autorid (d.h. die ID des Autors aus der Autorennamen-Tabelle)

 # @mail_references=();
 &get_message_references (@mail_header);	# set @mail_references
  
 if (!$message_inreplyto && scalar @mail_references == 0) {
    $message_first = 1;
    }
    else {				# check if a new thread is started
          $message_first = 0;
	  }
  
 &set_Autor ();		# insert new data in "Autor" table
    
 $message_subject = &get_message_subject (@mail_header);
 
 &set_Mail ();		# inser new date in "Mail" Table
          
 }	# end while there are new mails

close (MLARCHIVEFILE);

} # end sub mla_store


# pass2 is calculating threads and number of "last" E-Mails in a thread
# this will be stored then in the tables for further analysis

sub mla_pass2 {

my $mailid;		# numerical mail ID (table key)
my $mail_references;	# references of this E-Mail
my %thread;		# used to store references of threads
my %mail;		# to know if all mails are processed
my %maildate;		# stores date of mail(message-id)
my %mail_id;		# stores mail ids
my %mail_ref;		# for mail-ids: 1 has own references
my %mail_autor;		# author of mail
my $mailno;		# currently processed mail
my @refmails;		# referenced mails

$sql_command="select id, autorennamenid, messageid, mailreferences, maildate from Mail ORDER BY maildate";
$sql_sth = $dbh->prepare($sql_command);
$sql_rv = $sql_sth->execute();
if ($sql_sth->rows == 0) {
   print ("There seem to be no mails in the \"Mail table\"!\n");
   $sql_sth->finish;
   return();
  }
  else {
    
   while (($mailid, $autorennamen_autorid, $message_id, $mail_references, $message_date) = $sql_sth->fetchrow_array) {
   
       # print ("Mail gefunden: \$mailid, \$autorennamen_autorid, \$message_id, \$mail_references)= '$mailid', '$autorennamen_autorid', '$message_id', '$mail_references'\n");
       
       $maildate{$mailid} =$message_date; # store date of mail
       $mail_id{$message_id}=$mailid; # store id of mail
       $mail_autor{$mailid}=$autorennamen_autorid; # store author of mail
       $mail{$message_id} += 1;	# remember processed mails
       
       if ($mail_references) {
       # print ("Bongo... :-) '$mail_references'\n");
       $mail_ref{$mailid}=1;	# can't be start of thread, remember
       @mail_references = split ("\n", $mail_references);
       foreach (@mail_references) {
       # print ("Reference: '$_'\n");
       $thread{$_} .= "$mailid\n"; # add message-id to referenced id
       
       }   # end foreach
      } # end if references
      
   
   } # end while fetchrow

   print ("Mails found: \n");
    foreach (keys %mail) {
     $mailid=$mail_id{$_};
     print ("\nEs geht um die Mail-ID $mailid, Message-ID: '$_', gefunden $mail{$_} mal.\n");
     if ($mail{$_} > 1) {
       print ("Prfen: Mail '$_' zu oft gefunden $mail{$_} mal.\n");
      }
     if (!$thread{$_}) {
       print ("Die Mail '$_' ist nicht Bestandteil eines Threads\n");
       }
      else {
        if (!$mail_ref{$mailid}) {	# references in this mail?
        print ("Die Mail '$_' ist Bestandteil eines Threads, dazu gehren:\n");		# # no, must be beginning of thread
	@refmails = split ("\n", $thread{$_});
        foreach $mailno (@refmails) {
          print ("Mail no $mailno, Datum: $maildate{$mailno}\n"); 
        }
	# now store this in the tables
        $sql_command="insert into Thread (autorid, datumstart, datumend, numberofmails, startmailid, mailids) VALUES (?, ?, ?, ?, ?, ?)";
        # print ("Statement '$sql_command'\n");
        $sql_sth = $dbh->prepare($sql_command);
        $sql_rv = $sql_sth->execute ($mail_autor{$mailid}, $maildate{$mailid}, $maildate{$refmails[$#refmails]}, scalar @refmails, $mailid, join (", ",@refmails));
        $sql_sth->finish;

	
	    
       } 
       else {
         print ("Die Mail '$_' ist Bestandteil eines Threads, aber nicht der Start\n");
         } 
     }
          
      
   }
   
     
    if (0) {    
    print ("Threads found;\n");
    foreach (keys %thread) {
     if (! ($mailid=$mail_id{$_})) {
       print ("\nZensur links - Bezug auf eine nicht eingelesene Mail mit der Message-ID: $_\n");
       next;		# reference to mail which is not in DB
       }
     print ("\nMessage-ID ($mailid): $_ is referenced by\n");
     @refmails = split ("\n", $thread{$_});
     foreach $mailno (@refmails) {
          
      print ("Mail no $mailno, Datum: $maildate{$mailno}\n"); 
    } 
   }  # end for each keys thread
  } # end if threads found
 
 } 

}


#### build edge list
# start is very similar to pass2 calculating threads and number of "last" E-Mails in a thread
# this will be stored then in the tables for further analysis
# every thread will be seen as a "cluster" of communication, so 
# everybod involves is talking to everybody (not directed, thread
# starter included)

sub buildedgelist {

my $i;			# loop counters
my $j;
my $mailid;		# numerical mail ID (table key)
my $mail_references;	# references of this E-Mail
my %thread;		# used to store references of threads
my %mail;		# to know if all mails are processed (counts message_id)
my %maildate;		# stores date of mail(message-id)
my %mail_id;		# stores mail ids
my %mail_ref;		# for mail-ids: 1 has own references
my %mail_autor;		# author of mail
my $mailno;		# currently processed mail
my %matrix;		# matrix (fields: number of communication)
			# matrix{1}{2}=5: 5 Mails written from mail id 1 to mail id 2
my @refmails;		# referenced mails

$sql_command="select id, autorennamenid, messageid, mailreferences, maildate from Mail WHERE maildate >= ? and maildate <= ? ORDER BY maildate";
$sql_sth = $dbh->prepare($sql_command);
$sql_rv = $sql_sth->execute($mla_startdate, $mla_enddate);
if ($sql_sth->rows == 0) {
   print ("There seem to be no mails in the \"Mail table\"!\n");
   $sql_sth->finish;
   return();
  }
  else {
    
   while (($mailid, $autorennamen_autorid, $message_id, $mail_references, $message_date) = $sql_sth->fetchrow_array) {
   
        # print ("Mail gefunden: \$mailid, \$autorennamen_autorid, \$message_id, \$mail_references)= '$mailid', '$autorennamen_autorid', '$message_id', '$mail_references'\n");
       
       $maildate{$mailid} =$message_date; # store date of mail
       $mail_id{$message_id}=$mailid; # store id of mail
       $mail_autor{$mailid}=$autorennamen_autorid; # store author of mail
       $mail{$message_id} += 1;	# remember processed mails
       
       if ($mail_references) {
       # print ("Bongo... :-) '$mail_references'\n");
       $mail_ref{$mailid}=1;	# can't be start of thread, remember
       @mail_references = split ("\n", $mail_references);
       foreach (@mail_references) {
       # print ("Reference: '$_'\n");
       $thread{$_} .= "$mailid\n"; # add message-id to referenced id
       
       }   # end foreach
      } # end if references
      
   
   } # end while fetchrow

   print ("Mails found: \n");
    foreach (keys %mail) {
     $mailid=$mail_id{$_};
     print ("\nEs geht um die Mail-ID $mailid, Message-ID: '$_', gefunden $mail{$_} mal.\n");
     if ($mail{$_} > 1) {
       print ("Prfen: Mail '$_' zu oft gefunden $mail{$_} mal.\n");
      }
     if (!$thread{$_}) {
       print ("Die Mail '$_' ist nicht Bestandteil eines Threads - \$mail_ref ist $mail_ref{$mailid}\n");
       }
      else {
        if (!$mail_ref{$mailid}) {	# references in this mail?
        print ("Die Mail '$_' ist Bestandteil eines Threads, dazu gehren:\n");		# # no, must be beginning of thread
	@refmails = split ("\n", $thread{$_});
	push (@refmails, $mailid);	# sender is part of thread
	print ("alle: @refmails\n");
	for ($i= 0; $i<=$#refmails; $i++)
	  {
	   for ($j= 0; $j<=$#refmails; $j++)
	   {
	    # print ("i und j: $i, $j\n");
	    if ($i == $j) {
	     next;}		# person talks to itself
	   $matrix{$mail_autor{$refmails[$i]}}{$mail_autor{$refmails[$j]} } += 1;
	   print ("\$matrix{$mail_autor{$refmails[$i]}}{{$mail_autor{$refmails[$j]}} ist jetzt: $matrix{$mail_autor{$refmails[$i]}}{$mail_autor{$refmails[$j]}}\n");
	   } # end for j
	  } # end for i
       } 
       else {
         print ("Die Mail '$_' ist Bestandteil eines Threads, aber nicht der Start\n");
         } 
     } # end else if not thread
   } # end for each keys %mail
 
   print ("now writing edgelist\n");
   my @labels=("NULL");
   my $autorfullname;	# author full name out of DB table
   my $autoralias;	# author alias (if encoded in ISO-8859)
   my $edgelistline;	# line of edge list to be stored
   my $numberofcomm;	# number of communications between to authors
   my @edgelist=("data:\n");	# edgelist (holds edgelistlines)
   # collect names
   my $ic=1;	# counter of i [1 .. n], wheres i counts author IDs!
   my $jc=1;	# counter of j, same thing
   
   foreach $i (sort keys %matrix) {
         
	 print ("found authorid $i\n");
	 $sql_command="select autorname, autoralias from Autorennamen where autorid = ?";
	 $sql_sth = $dbh->prepare($sql_command);
	 $sql_rv = $sql_sth->execute($i);
	 if ($sql_sth->rows == 0)
  	  {
   	   print ("Den Autor mit der ID $i gibt es nicht!\n");
   	   die ("Da kann etwas nicht stimmen!\n");
	  }
   	 else {
	   ($autorfullname, $autoralias)=$sql_sth->fetchrow;
	   if ($autoralias && ($autorfullname =~ /\=iso/)) {
	    $autorfullname=$autoralias;	# use alias if encoded fullname
	    }
	   print ("Autor mit der ID $i ist '$autorfullname'\n");
	   $_ = $autorfullname;
	   if (/=\?iso/i) {
	    $_=unmime ($_);
	    print ("decoded: $_ (was before $autorfullname)\n");
	    }
	   s/^,//;	# remove "," at beginning
	   s/\s/./g;	# replace space with "."
	   s/,/_/g;	# replace , with "_"
	   s/\(/_/g;	# replace ( with "_"
	   s/\)/_/g;	# replace ) with "_"
	   push (@labels, "$ic-$_");
           $jc =1;	# reset edge list counter
	   foreach $j (sort keys %matrix) {
	   # print ("\$j ist '$j'\n");
	   if ($i == $j) {
	      $jc++;
	      next;
	      }
	   if ($matrix{$i}{$j}) {
	     $numberofcomm = $matrix{$i}{$j};
	     }
	     else {
	     $numberofcomm = 0;
	     }
	   $edgelistline=$ic . " " . $jc . " " . $numberofcomm. "\n";
	   print ("$edgelistline\n");
	   push (@edgelist, $edgelistline);   
	   $jc++;
          } # end foreach $j
	 }  # end else author in table
	 $ic++;
 	 }  # end foreach $i
    
    $sql_sth->finish;
      
   if (!(open FILEOUTL, ">$edgelistfile")) {
       die (" Can't open $edgelistfile: $!\n");
       };
   if (!(print FILEOUTL "dl n=$#labels\nformat = edgelist1\nlabels:\n"))
      {
       die ("Can't write to $edgelistfile $!\n");
      };
   for ($i=1; $i <= $#labels; $i++) {
     print FILEOUTL "\t$labels[$i]\n";
   }
 print FILEOUTL "@edgelist\n";     
         
 close (FILEOUTL);
    } 
    
} # end sub
   

 
  






###############################################################
##############Programmstart#######################################

@mlarchiveprefix=();		# reset archived Mail
$mlarchivefilelineno="1";	# line counter - begin with 1

my $numberofemailinfilecounter=0;	# number of email in file
my $lineinfileofcurrentemail=1;	# line of Email is it right now

use DBI;

my $datenbank="kdedevel";   # Name der Datenbank
my $datenbankhost="localhost";  # Rechner, auf dem die Datenbank luft
my $datenbankuser="dbroot";   # Benutzername und Passwort
my $datenbankpw="geheim";

# Verbindung zur Datenbank herstellen

$dbh=DBI->connect("DBI:mysql:$datenbank:$datenbankhost","$datenbankuser",  "$datenbankpw") || die ("Fehler bei der Datenbankverbindung zu $datenbank, $DBI::errstr\n");

# haut nicht hin: ---   , {RaiseError -> 1}

# Datei wird von der Kommandozeile bergeben

if (&get_options == -1) {
  print ("Usage: $0 -store mailinglistfile (parses file with ML-Emails and stores them in Database)\n");
  print "or: $0 -pass2 (builds new fields after all mails have been stored)\n";
  print "or: $0 -buildedgelist [-startdate=\"date\"] [-enddate=\"date\"] (extracts results for network analysis)\n";
  die();
  } 
  
if ($mla_mode eq "store")
  {
   foreach (@filenamelist) {
    $mlarchivefile = $_;
    &mla_store();
    }
  }

  elsif ($mla_mode eq "pass2") {
   &mla_pass2();
  }
  elsif ($mla_mode eq "buildedgelist") {
   &buildedgelist();
   }

  
  
$dbh->disconnect; # Verbindung zur DB beenden

die();

#########

my $line;
foreach $line (@mail_body)
  {
  print ("Zeile: '$line'\n");
  if (!defined $line) {
   print ("HIERHIERHIER\n\n\n");
   }
  }
